@@ -50,9 +50,9 @@ __all__ = ('BaseDependency', 'Dependency', 'Origin', 'Package', 'Record',
 
 def _file_is_same(path, size, md5):
     """Return ``True`` if the file is the same."""
-    if (os.path.exists(path) and os.path.getsize(path) == size and
-        apt_pkg.md5sum(open(path)) == md5):
-        return True
+    if os.path.exists(path) and os.path.getsize(path) == size:
+        with open(path) as fobj:
+            return apt_pkg.md5sum(fobj) == md5
 
 
 class FetchError(Exception):
@@ -94,7 +94,7 @@ class BaseDependency(object):
         preDepend = AttributeDeprecatedBy('pre_depend')
 
 
-class Dependency(object):
+class Dependency(list):
     """Represent an Or-group of dependencies.
 
     Attributes defined here:
@@ -102,11 +102,12 @@ class Dependency(object):
     """
 
     def __init__(self, alternatives):
-        self.or_dependencies = alternatives
-
-    def __repr__(self):
-        return repr(self.or_dependencies)
+        super(Dependency, self).__init__()
+        self.extend(alternatives)
 
+    @property
+    def or_dependencies(self):
+        return self
 
 class DeprecatedProperty(property):
     """A property which gives DeprecationWarning on access.
@@ -120,7 +121,7 @@ class DeprecatedProperty(property):
         self.__doc__ = (doc or fget.__doc__ or '')
 
     def __get__(self, obj, type_=None):
-        if obj is not None and "PYTHON_APT_DEPRECATION_WARNINGS" in os.environ:
+        if obj is not None:
             warnings.warn("Accessed deprecated property %s.%s, please see the "
                           "Version class for alternatives." %
                            ((obj.__class__.__name__ or type_.__name__),
@@ -162,10 +163,23 @@ class Origin(object):
 
 
 class Record(Mapping):
-    """Represent a pkgRecord.
+    """Record in a Packages file
+
+    Represent a record as stored in a Packages file. You can use this like
+    a dictionary mapping the field names of the record to their values::
+
+        >>> record = Record("Package: python-apt\\nVersion: 0.8.0\\n\\n")
+        >>> record["Package"]
+        'python-apt'
+        >>> record["Version"]
+        '0.8.0'
+
+    For example, to get the tasks of a package from a cache, you could do::
+
+        package.candidate.record["Tasks"].split()
+
+    Of course, you can also use the :attr:`Version.tasks` property.
 
-    It can be accessed like a dictionary and can also give the original package
-    record if accessed as a string.
     """
 
     def __init__(self, record_str):
@@ -209,6 +223,9 @@ class Record(Mapping):
 class Version(object):
     """Representation of a package version.
 
+    The Version class contains all information related to a
+    specific package version.
+
     .. versionadded:: 0.7.9
     """
 
@@ -327,7 +344,7 @@ class Version(object):
     def description(self):
         """Return the formatted long description.
 
-        Return the formated long description according to the Debian policy
+        Return the formatted long description according to the Debian policy
         (Chapter 5.6.13).
         See http://www.debian.org/doc/debian-policy/ch-controlfields.html
         for more information.
@@ -338,7 +355,7 @@ class Version(object):
             if not isinstance(dsc, unicode):
                 # Only convert where needed (i.e. Python 2.X)
                 dsc = unicode(dsc, "utf-8")
-        except UnicodeDecodeError, err:
+        except UnicodeDecodeError as err:
             return _("Invalid unicode in description for '%s' (%s). "
                   "Please report.") % (self.package.name, err)
 
@@ -374,18 +391,41 @@ class Version(object):
     def source_name(self):
         """Return the name of the source package."""
         try:
-            return self._records.source_pkg or self.package.name
+            return self._records.source_pkg or self.package.shortname
         except IndexError:
-            return self.package.name
+            return self.package.shortname
+
+    @property
+    def source_version(self):
+        """Return the version of the source package."""
+        try:
+            return self._records.source_ver or self._cand.ver_str
+        except IndexError:
+            return self._cand.ver_str
 
     @property
     def priority(self):
         """Return the priority of the package, as string."""
         return self._cand.priority_str
 
+    @property
+    def policy_priority(self):
+        """Return the internal policy priority as a number.
+           See apt_preferences(5) for more information about what it means.
+        """
+        priority = 0
+        policy = self.package._pcache._depcache.policy
+        for (packagefile, _) in self._cand.file_list:
+            priority = max(priority, policy.get_priority(packagefile))
+        return priority
+
     @property
     def record(self):
-        """Return a Record() object for this version."""
+        """Return a Record() object for this version.
+
+        Return a Record() object for this version which provides access
+        to the raw attributes of the candidate version
+        """
         return Record(self._records.record)
 
     def get_dependencies(self, *types):
@@ -406,6 +446,11 @@ class Version(object):
                 pass
         return depends_list
 
+    @property
+    def provides(self):
+        """ Return a list of names that this version provides."""
+        return [p[0] for p in self._cand.provides_list]
+        
     @property
     def enhances(self):
         """Return the list of enhances for the package version."""
@@ -421,6 +466,11 @@ class Version(object):
         """Return the recommends of the package version."""
         return self.get_dependencies("Recommends")
 
+    @property
+    def suggests(self):
+        """Return the suggests of the package version."""
+        return self.get_dependencies("Suggests")
+
     @property
     def origins(self):
         """Return a list of origins for the package version."""
@@ -461,6 +511,16 @@ class Version(object):
         """
         return self._records.sha256_hash
 
+    @property
+    def tasks(self):
+        """Get the tasks of the package.
+
+        A set of the names of the tasks this package belongs to.
+
+        .. versionadded:: 0.8.0
+        """
+        return set(self.record["Task"].split())
+
     def _uris(self):
         """Return an iterator over all available urls.
 
@@ -485,7 +545,10 @@ class Version(object):
 
         .. versionadded:: 0.7.10
         """
-        return iter(self._uris()).next()
+        try:
+            return iter(self._uris()).next()
+        except StopIteration:
+            return None
 
     def fetch_binary(self, destdir='', progress=None):
         """Fetch the binary version of the package.
@@ -502,8 +565,8 @@ class Version(object):
         base = os.path.basename(self._records.filename)
         destfile = os.path.join(destdir, base)
         if _file_is_same(destfile, self.size, self._records.md5_hash):
-            print 'Ignoring already existing file:', destfile
-            return
+            print('Ignoring already existing file: %s' % destfile)
+            return os.path.abspath(destfile)
         acq = apt_pkg.Acquire(progress or apt.progress.text.AcquireProgress())
         acqfile = apt_pkg.AcquireFile(acq, self.uri, self._records.md5_hash,
                                       self.size, base, destfile=destfile)
@@ -512,7 +575,7 @@ class Version(object):
         if acqfile.status != acqfile.STAT_DONE:
             raise FetchError("The item %r could not be fetched: %s" %
                              (acqfile.destfile, acqfile.error_text))
-        print self._records.filename
+
         return os.path.abspath(destfile)
 
     def fetch_source(self, destdir="", progress=None, unpack=True):
@@ -536,7 +599,7 @@ class Version(object):
 
         dsc = None
         record = self._records
-        source_name = record.source_pkg or self.package.name
+        source_name = record.source_pkg or self.package.shortname
         source_version = record.source_ver or self._cand.ver_str
         source_lookup = src.lookup(source_name)
 
@@ -551,7 +614,7 @@ class Version(object):
             if type_ == 'dsc':
                 dsc = destfile
             if _file_is_same(destfile, size, md5):
-                print 'Ignoring already existing file:', destfile
+                print('Ignoring already existing file: %s' % destfile)
                 continue
             files.append(apt_pkg.AcquireFile(acq, src.index.archive_uri(path),
                          md5, size, base, destfile=destfile))
@@ -661,7 +724,11 @@ class Package(object):
         self._changelog = ""            # Cached changelog
 
     def __repr__(self):
-        return '<Package: name:%r id:%r>' % (self._pkg.name, self._pkg.id)
+        return '<Package: name:%r architecture=%r id:%r>' % (self._pkg.name,
+                 self._pkg.architecture, self._pkg.id)
+
+    def __lt__(self, other):
+        return self.name < other.name
 
     def candidate(self):
         """Return the candidate version of the package.
@@ -693,7 +760,30 @@ class Package(object):
 
     @property
     def name(self):
-        """Return the name of the package."""
+        """Return the name of the package, possibly including architecture.
+
+        If the package is not part of the system's preferred architecture,
+        return the same as :attr:`fullname`, otherwise return the same
+        as :attr:`shortname`
+
+        .. versionchanged:: 0.7.100.3
+        As part of multi-arch, this field now may include architecture
+        information.
+        """
+        return self._pkg.get_fullname(True)
+
+    @property
+    def fullname(self):
+        """Return the name of the package, including architecture.
+
+        .. versionadded:: 0.7.100.3"""
+        return self._pkg.get_fullname(False)
+
+    @property
+    def shortname(self):
+        """Return the name of the package, without architecture.
+
+        .. versionadded:: 0.7.100.3"""
         return self._pkg.name
 
     @property
@@ -744,13 +834,15 @@ class Package(object):
         """
         return getattr(self.installed, 'dependencies', [])
 
-    @DeprecatedProperty
     def architecture(self):
         """Return the Architecture of the package.
 
-        .. deprecated:: 0.7.9
+        .. versionchanged:: 0.7.100.3
+            This is now the package's architecture in the multi-arch sense,
+            previously it was the architecture of the candidate version
+            and deprecated.
         """
-        return getattr(self.candidate, "architecture", None)
+        return self._pkg.architecture
 
     @DeprecatedProperty
     def candidateDownloadable(self):  #pylint: disable-msg=C0103
@@ -823,7 +915,7 @@ class Package(object):
     def description(self):
         """Return the formatted long description.
 
-        Return the formated long description according to the Debian policy
+        Return the formatted long description according to the Debian policy
         (Chapter 5.6.13).
         See http://www.debian.org/doc/debian-policy/ch-controlfields.html
         for more information.
@@ -904,8 +996,8 @@ class Package(object):
         another package, and if no packages depend on it anymore, the package
         is no longer required.
         """
-        return self.is_installed and \
-               self._pcache._depcache.is_garbage(self._pkg)
+        return ((self.is_installed or self.marked_install) and
+                self._pcache._depcache.is_garbage(self._pkg))
 
     @property
     def is_auto_installed(self):
@@ -955,11 +1047,8 @@ class Package(object):
         """
         path = "/var/lib/dpkg/info/%s.list" % self.name
         try:
-            file_list = open(path, "rb")
-            try:
+            with open(path, "rb") as file_list:
                 return file_list.read().decode("utf-8").split(u"\n")
-            finally:
-                file_list.close()
         except EnvironmentError:
             return []
 
@@ -1006,34 +1095,31 @@ class Package(object):
         src_section = "main"
         # use the section of the candidate as a starting point
         section = self.candidate.section
-
-        # get the source version, start with the binaries version
-        bin_ver = self.candidate.version
-        src_ver = self.candidate.version
-        #print "bin: %s" % binver
+        
+        # get the source version
+        src_ver = self.candidate.source_version
+                
         try:
-            # FIXME: This try-statement is too long ...
             # try to get the source version of the pkg, this differs
             # for some (e.g. libnspr4 on ubuntu)
             # this feature only works if the correct deb-src are in the
-            # sources.list
-            # otherwise we fall back to the binary version number
+            # sources.list otherwise we fall back to the binary version number
             src_records = apt_pkg.SourceRecords()
-            src_rec = src_records.lookup(src_pkg)
-            if src_rec:
-                src_ver = src_records.version
-                #if apt_pkg.VersionCompare(binver, srcver) > 0:
-                #    srcver = binver
-                if not src_ver:
-                    src_ver = bin_ver
-                #print "srcver: %s" % src_ver
-                section = src_records.section
-                #print "srcsect: %s" % section
-            else:
-                # fail into the error handler
-                raise SystemError
         except SystemError:
-            src_ver = bin_ver
+            pass
+        else:
+            while src_records.lookup(src_pkg):
+                if not src_records.version:
+                    continue
+                if self.candidate.source_version == src_records.version:
+                    # Direct match, use it and do not do more lookups.
+                    src_ver = src_records.version
+                    section = src_records.section
+                    break
+                if apt_pkg.version_compare(src_records.version, src_ver) > 0:
+                    # The version is higher, it seems to match.
+                    src_ver = src_records.version
+                    section = src_records.section
 
         section_split = section.split("/", 1)
         if len(section_split) > 1:
@@ -1068,6 +1154,7 @@ class Package(object):
                 # Check if the download was canceled
                 if cancel_lock and cancel_lock.isSet():
                     return u""
+                # FIXME: python3.2: Should be closed manually
                 changelog_file = urllib2.urlopen(uri)
                 # do only get the lines that are new
                 changelog = u""
@@ -1078,7 +1165,7 @@ class Package(object):
                         return u""
                     # Read changelog line by line
                     line_raw = changelog_file.readline()
-                    if line_raw == "":
+                    if not line_raw:
                         break
                     # The changelog is encoded in utf-8, but since there isn't
                     # any http header, urllib2 seems to treat it as ascii
@@ -1160,7 +1247,7 @@ class Package(object):
 
     @deprecated_args
     def mark_delete(self, auto_fix=True, purge=False):
-        """Mark a package for install.
+        """Mark a package for deletion.
 
         If *auto_fix* is ``True``, the resolver will be run, trying to fix
         broken packages.  This is the default.
@@ -1296,8 +1383,8 @@ def _test():
     print "SourcePkg: %s " % pkg.candidate.source_name
     print "Section: %s " % pkg.section
     print "Summary: %s" % pkg.candidate.summary
-    print "Description (formated) :\n%s" % pkg.candidate.description
-    print "Description (unformated):\n%s" % pkg.candidate.raw_description
+    print "Description (formatted) :\n%s" % pkg.candidate.description
+    print "Description (unformatted):\n%s" % pkg.candidate.raw_description
     print "InstalledSize: %s " % pkg.candidate.installed_size
     print "PackageSize: %s " % pkg.candidate.size
     print "Dependencies: %s" % pkg.installed.dependencies
